# -*- coding: utf-8 -*-

# Script to automatically verify FOO with weighted votes
# by Jannik Dreier, jannik.dreier@imag.fr , 2013
#
# Parameters:
# - Minimal weight "mini",
# - Maximal weight "maxi",
# - number of voters "voters",
# - sum of all weights "tar"

import subprocess
import sys

def processV(file, id, n):
  """ function that generates the "processV" for ID id and n votes. The text is written to file. """
  file.write("let processV%d = \n" % (id))
  for k in range(n):
    file.write("  new b%d ; new r%d ;\n" % (k, k))
    file.write("  let commitedvote%d = commit(v, r%d) in\n" % (k, k))
    file.write("  let blindedvote%d = blind(commitedvote%d, b%d) in\n" % (k, k, k))
    file.write("  out(ch1, (pk(skv), sign(blindedvote%d, skv), blindedvote%d));\n" % (k, k))
  for k in range(n):
    file.write("  in(ch2, m%d);\n" % k)
    file.write("  let result%d = checksign(m%d, pk(ska)) in\n" % (k, k))
    file.write("  if result%d = blindedvote%d then\n" % (k, k))
    file.write("  let signedvote%d = unblind(m%d, b%d) in\n" % (k, k, k))
  file.write("  sync 1;\n")
  for k in range(n):
    file.write("  out(ch3, (commitedvote%d, signedvote%d));\n" % (k, k))
    file.write("  in(ch4, (l%d, =commitedvote%d, =signedvote%d));\n" % (k, k, k))
  file.write("  sync 2;\n")
  for k in range(n):
    file.write("  out(ch5, (l%d, r%d));\n" % (k, k))
  file.write("  0.\n")

def toLetter(k):
  """ Transforms a number into a letter, wehere 0 = a """
  return chr(k+97)

def printPI(votecount, votea, voteb, ws, vas, vbs):
  """ Generates the Applied-PI-Code, where votecount is a list containing the weights, votea and voteb are arrays containing the choices of each voter and ws, vas and vbs are strings containing the weights and the votes in case a and c respectively (used only for the filename) """
  file = open("foow/%s-%s-%s.piswap" % (ws, vas, vbs), "w")
  file.write("""free ch, ch1, ch2, ch3, ch4, ch5, c1, c2.
free a, b, v.
free ska, skva, skvb, skvc, pka.
free skvaCh, skvbCh, skvcCh.

fun open/2.
fun commit/2.
fun checksign/2.
fun sign/2.
fun blind/2.
fun unblind/2.
fun pk/1.
fun true/0.

equation open(commit(m, r), r) = m.
equation checksign(sign(m, sk), pk(sk)) = m.
equation unblind(blind(m, r), r) =  m.
equation unblind(sign(blind(m, r), sk), r) = sign(m, sk).
""")

  for k in range(len(votecount)):
	 processV(file, k, votecount[k])
  
  file.write("\n")
  
  #processVc1c2(file, m)
  #votes[k]
  file.write("""
  
process
(""")
  for k in range(len(votecount)-1):
	 file.write(" (let skv = skv%c in let v = choice[%c, %c] in processV%d) |\n" % (toLetter(k), toLetter(votea[k]), toLetter(voteb[k]), k))
  file.write(" (let skv = skv%c in let v = choice[%c, %c] in processV%d)\n" % (toLetter(len(votecount)-1), toLetter(votea[len(votecount)-1]), toLetter(voteb[len(votecount)-1]), len(votecount)-1))
  file.write(""" )
""")

  file.close

def combinations(mini, maxi, voters, l, ret):
  """ adds all possible combinations of vote choices for "voters" voters to "ret", "l" is a list that is used as accumulator to construct the values """
  for k in range(mini, maxi+1):
	 l.append(k)
	 if voters > 1: 
		combinations(mini, maxi, voters-1, l, ret)
	 else:
		ret.append(list(l))
	 l.pop()

def weights(mini, maxi, voters, l, su, tar, ret):
  """ calculates all possible combinations of weights for "voters" voters, where the weights are > mini and < maxi, and the sum of all weights has to be equal to "tar"
  "l" (list) and "su" (sum) are recursion variables, "ret" returns the list of possible weight distributions """
  if voters > 1: 
	 for k in range(mini, maxi+1):
		l.append(k)
		if (su + k) < tar:
		  weights(k, maxi, voters-1, l, su + k, tar, ret)
		l.pop()
  else:
	 k = tar - su;
	 if k >= mini and k <= maxi:
		l.append(k)
		ret.append(list(l))
		l.pop()

def access (ar, res):
  """ helper method to access data inside a multidimensional array "ar" where the coordinates are given as a list "res" """
  if len(res) == 1:
	 return ar[res[0]]
  else:
	 return access (ar[res[0]], res[1:]) 

def insert (ar, res, v):
  """ helper method to insert a vote "v" inside a multidimensional array "ar" where the coordinates are given as a list "res"
  return True if there is a collision (another entry was made before), False otherwise """
  if len(res) == 1:
	 if ar[res[0]] == []:
		ar[res[0]] = [v]
		return False
	 else:
	   ar[res[0]].append(v)
	   return True
  else:
	 return insert (ar[res[0]], res[1:], v) 

def calculateResult (w, v):
  """ calculates the result for a weight distribution "w" and a vote distribution "v" """
  res = [0 for col in range(len(v))]
  for k in range(0, len(w)):
	 res[v[k]]+=w[k]
  return res

def initar(k, n):
  """ creates an array with "k" dimensions of "n" empty lists each """
  if k == 1:
	 return [[] for col in range(n)]
  else:
	 return [initar(k-1, n) for col in range(n)]

def elimDoubles(l):
  """ eleminates double entries in a list "l" """
  res = []
  for e in l:
	 if res.count(e) == 0:
		res.append(e)
  return res

def arToString(ar):
  """ Turns a list "ar" into a string, e.g. [1, 2] to 1_2
  used to generate file names """
  s = ""
  first = True
  for k in ar:
	 if first:
		s = str(k)
		first = False
	 else:
		s += "_" + str(k)
  return s

def check (w, combs, ar):
  """ checks for bisimlarity of the protocol
  "w": weight distribution, "combs": List of results that have more than one entry (i.e. more than one possibility to be obtained) in the array "ar" """
  for c in combs:
	 v = access(ar, c)
	 for k in range(len(v)):
		for l in range(k+1, len(v)):
		  ws = arToString(w)
		  vas = arToString(v[k])
		  vbs = arToString(v[l])
		  printPI(w, v[k], v[l], ws, vas, vbs)
		  output = open("foow/%s-%s-%s.pi" % (ws, vas, vbs), "w")
		  proswapper = subprocess.Popen(["proswapper", "foow/%s-%s-%s.piswap" % (ws, vas, vbs)], stdout=output)
		  proswapper.wait()
		  #output.flush()
		  output.close()
		  output = file("foow/%s-%s-%s.pi" % (ws, vas, vbs), "r")
		  newlines = []
		  for line in output:
			 if "| sync_process_0" in line:
				line = "| sync_process_2"
			 newlines.append(line)
		  output.close()
		  outfile = file("foow/%s-%s-%s.pi" % (ws, vas, vbs), "w")
		  outfile.writelines(newlines)
		  outfile.close()
		  outverif = open("foow/%s-%s-%s.verif" % (ws, vas, vbs), "w")
		  proverif = subprocess.Popen(["proverif", "-in", "pi", "foow/%s-%s-%s.pi" % (ws, vas, vbs)], stdout=outverif)
		  proverif.wait()
		  outverif.close()
		  outverif = open("foow/%s-%s-%s.verif" % (ws, vas, vbs), "r")
		  tofind = 'RESULT Observational equivalence is true'
		  sucess = False
		  for line in outverif:
			 if tofind in line:
				sucess = True
		  if sucess:
			 print "Case %s-%s-%s: OK" % (ws, vas, vbs)
		  else:
			 print "Case %s-%s-%s: KO" % (ws, vas, vbs)
		  outverif.close()


def calculateCases(mini, maxi, voters, tar):
  """ calculates the cases where two instances give the same result and checks them for bisimilarity
   Parameters: Minimal weight "mini", Maximal weight "maxi", number of voters "voters", sum of all weights "tar"
   typically tar = maxi + (voters-1) """
  weight = []
  weights(mini, maxi, voters, [], 0, tar, weight)
  votes = []
  combinations(0, voters-1, voters, [], votes)
  for w in weight:
	 combs = []
	 ar = initar(voters, tar+1)
	 for v in votes:
		res = calculateResult(w, v)
		resc = res[:]
		if insert(ar, res, v):
		  combs.append(resc)
	 combs = elimDoubles(combs)
	 check(w, combs, ar)


# main function
if __name__ == '__main__':
  calculateCases(1, 3, 3, 5) 
  #sys.exit(0)