# -*- coding: utf-8 -*-
import subprocess
import sys

# Generates ProVerif input to check Weak Noninterference for two bidders and two prices, and launches ProVerif on the resulting files.
# Files are stored in directory sako-wni and named e.g. 1-1-1-1.pi (for the input) and 1-1-1-1.verif (for the output)

def getString(number):
    if number==0:
        return "empty"
    if number==1:
        return "one"
    if number==2:
        return "two"
    if number==3:
        return "three"
    if number==4:
        return "four"
    if number==5:
        return "five"
    if number==6:
        return "six"
    if number==7:
        return "seven"
    if number==8:
        return "eight"
    if number==9:
        return "nine"
    if number==10:
        return "ten"
    else:
        return "error"

def printCode(file):
    file.write("""(* sako.pi : protocol by Sako (2000) *)
(* Property: Weak Non-Interference *)

type skey.
type pkey.
type bid.
type signature.
type int.
type seed.

event send(bitstring, bitstring).
event OK().
event KO().

free chBB1:channel.
free chBB2:channel.
free chPKI:channel.
free chAd:channel.

fun pubkey(skey):pkey.
fun sign(bitstring, skey):bitstring.
fun enc(bitstring, pkey, seed):bitstring.
fun dec(bitstring, skey):bitstring.
fun one():int.
fun two():int.
fun three():int.
fun suc(int):int.
fun geq(int, int):bool.
fun intToBit(int):bitstring.
fun bidval():bitstring.
fun decProof(bitstring,bitstring,skey):bitstring.
fun checkproof(bitstring,bitstring,bitstring,pkey):bool.

reduc forall m:bitstring, k:skey; checksign(sign(m,k),pubkey(k)) = m.
reduc forall m:bitstring, k:skey; getmessage(sign(m,k)) = m.

equation forall m:bitstring, k:skey, s:seed; dec(enc(m,pubkey(k),s),k) = m.

equation forall m:bitstring, k:skey, s:seed; checkproof(decProof(enc(m, pubkey(k), s),m,k),enc(m, pubkey(k), s),m,pubkey(k)) = true.
equation forall m:bitstring, k1:skey, k2:skey, s:seed; checkproof(decProof(enc(m, pubkey(k1), s),dec(enc(m, pubkey(k1), s),k2),k2),enc(m, pubkey(k1), s),dec(enc(m, pubkey(k1), s),k2),pubkey(k2)) = true.

let bidder(b:pkey, k:skey, chBB:channel) =
  new s:seed;
  let offer:bitstring = enc(bidval(), b, s) in
  event send(offer, sign(offer, k));
  out(chBB, (offer, sign(offer, k))).

let instance(b1:pkey, b2:pkey) =
 ((new k:skey; out(chPKI, pubkey(k)); bidder(b1, k, chBB1))) | 
 ((new k:skey; out(chPKI, pubkey(k)); bidder(b2, k, chBB2))).

process
 new keyone:skey; new keytwo:skey;
 ( out(chPKI, pubkey(keyone)) |
   out(chPKI, pubkey(keytwo)) |
   out(chPKI, one) |
   out(chPKI, two) |""")

  
# main function
if __name__ == '__main__':
    for i in range(1,3):
        for j in range(1,3):
            for k in range(1,3):
                for l in range(1,3):
                    output = open("sako-wni/%d-%d-%d-%d.pi" % (i, j, k, l), "w")
                    printCode(output)
                    output.write("""(instance(choice[pubkey(key%s),pubkey(key%s)],choice[pubkey(key%s),pubkey(key%s)])) )\n""" % (getString(i), getString(j), getString(k), getString(l)))
                    output.close()
                    outverif = open("sako-wni/%d-%d-%d-%d.verif" % (i, j, k, l), "w")
                    proverif = subprocess.Popen(["proverif", "-in", "pitype", "sako-wni/%d-%d-%d-%d.pi" % (i, j, k, l)], stdout=outverif)
                    proverif.wait()
                    outverif.close()
                    outverif = open("sako-wni/%d-%d-%d-%d.verif" % (i, j, k, l), "r")
                    tofind = 'RESULT Observational equivalence is true'
                    sucess = False
                    for line in outverif:
                      if tofind in line:
                          sucess = True
                    if sucess:
                      print ("Case %d-%d-%d-%d: OK" % (i, j, k, l))
                    else:
                      print ("Case %d-%d-%d-%d: KO" % (i, j, k, l))
                    outverif.close()
    #sys.exit(0)
